import threading
from multiprocessing import cpu_count
from os.path import exists, join
from numpy import arange
from pandas import DataFrame, concat

try:
    from functions import *
    from utils import split_test_groups, convert_str_to_day
    from configs import hyper_conf, descriptive_columns
except Exception as e:
    from .functions import *
    from .utils import split_test_groups, convert_str_to_day
    from .configs import hyper_conf, descriptive_columns


def get_confidence_intervals(intervals):
    return np.arange(intervals[0], intervals[1], intervals[2])


def get_comb_params(distributions):
    comb_arrays = {}
    for d in distributions:
        _params = distributions[d]
        print(_params)
        _keys = list(distributions[d].keys())
        arrays = []
        for p in _params:
            print(p)
            if "*" in list(_params[p]):
                arrays.append(
                              arange(float(_params[p].split("*")[0]),
                                     float(_params[p].split("*")[1]),
                                     float(_params[p].split("*")[0])).tolist()
                    )
            else:
                arrays.append([float(i) for i in _params[p].split("_")])
        comb_arrays[d] = list(product(*arrays))
    return comb_arrays


def is_numeric_value_an_integer(value):
    if str(float(value)).split(".")[-1] == '0':
        return int
    else:
        return float


def get_params(keys, comb, data):
    count, params = 0, {}
    for p in keys:
        _p = is_numeric_value_an_integer(comb[count])(comb[count])
        params[p] = _p
        count += 1
    if len(data) < 300:
        params['iteration'] = 1
        params['sample_size'] = 1
    return params


def rename_descriptives():
    d = {}
    for i in descriptive_columns:
        d[i] = i[:-1] + '_control' if i[-1] == '1' else i[:-1] + '_validation'
    return d


def assign_groups_to_results(data, groups, comb):
    if len(groups) != 0:
        count = 0
        for g in groups:
            data[g] = comb[count]
    return data


class Test:
    def __init__(self,
                 test_groups,
                 groups=None,
                 date=None,
                 feature=None,
                 data_source=None,
                 data_query_path=None,
                 time_period=None,
                 time_indicator=None,
                 export_path=None,
                 parameters=None):
        self.date = convert_str_to_day(date)
        self.time_indicator = time_indicator
        self.data, self.groups = data_manipulation(date=date,
                                                   time_indicator=time_indicator,
                                                   feature=feature,
                                                   data_source=data_source,
                                                   groups=groups,
                                                   data_query_path=data_query_path,
                                                   time_period=time_period)
        self.test_groups_field = test_groups
        self.test_groups_indicator = split_test_groups(self.test_groups_field, self.data)
        self.feature = feature
        self.time_period = time_period
        self.time_indicator_values = None
        self.tp = None
        self.levels = get_levels(self.data, self.groups)
        self._c, self._a = None, None
        self.f_w_data = None
        self.data_distribution = 'normal'  # by default it is Normal distribution
        self.parameter_combinations = get_comb_params(hyper_conf('distribution_parameters') if parameters is None else parameters)
        self.export_path = export_path
        self.comb, self.param_comb, self._params = None, None, {}
        self.results = []
        self.final_results = DataFrame()
        self.h0_accept_ratio = 0
        self.h0_acceptance = 0

    def get_query(self):
        count = 0
        query = ''
        for c in self.comb:
            if type(c) != str:
                query += self.groups[count] + ' == ' + str(c) + ' and '
            else:
                query += self.groups[count] + " == '" + str(c) + "' and "
            count += 1

        if not self.get_time_period_of_time_line():
            if self.time_period == 'day':
                print("yessssss")
                query += self.time_indicator + " <= '" + str(self.tp) + "' and "
            else:
                if self.time_period in ["year", "month", "week", "week_day", "hour"]:
                    query += self.time_period + " == " + str(self.tp) + " and "
                if self.time_period in ["quarter", "week_part", "day_part"]:
                    query += self.time + " == '" + str(self.tp) + "' and "
        query = query[:-4]
        print(query)
        return query

    def get_control_and_active_data(self):
        self._c = self.f_w_data[self.f_w_data[self.test_groups_field] == self.test_groups_indicator]
        self._a = self.f_w_data[self.f_w_data[self.test_groups_field] != self.test_groups_indicator]

    def decide_distribution(self):
        _unique = list(self.data[self.feature].unique())
        _type = type(_unique[0])
        # by default it is Normal distribution
        if _type != str:
            _min, _max = min(self.data[self.feature]), max(self.data[self.feature])
            if 0 <= _min < 1 and 0 < _max <= 1:
                self.data_distribution = 'beta'
        if len(_unique) == 2:
            self.data_distribution = 'binominal'
        if 2 < len(_unique) < 30:
            if _type == int:
                if min(self.data[self.feature]) >= 0:
                    self.data_distribution = 'poisson'
            if _type == str:
                self.data_distribution = 'poisson'
        print("Distribution :", self.data_distribution)

    def get_descriptives(self):
        self.results = self.results.rename(columns=rename_descriptives())

    def test_decision(self):
        if self.is_boostraping_calculation():
            self.get_descriptives()
            self.h0_accept_ratio = sum(self.results['h0_accept']) / len(self.results)
            self.h0_acceptance = True if self.h0_accept_ratio > 0.5 else False
            self.results['date'] = self.date
            self.results['test_result'] = self.h0_acceptance
            self.results['accept_Ratio'] = self.h0_accept_ratio
            self.results = assign_groups_to_results(self.results, self.groups, self.comb)
            self.final_results = self.results if self.final_results is None else concat([self.final_results, self.results])
        else:
            print(self.results.head())
            self.final_results = self.results

    def is_boostraping_calculation(self):
        if self.date is None:
            return True
        else:
            if self.export_path is None:
                return True
            else:
                files = check_result_data_exits(self.export_path)
                if len(files) >= 1:
                    return True
                else:
                    return False

    def test_execute(self):
        self.results = []
        for self.param_comb in self.parameter_combinations[self.data_distribution]:
            self._params = get_params(list(hyper_conf('normal').keys()), self.param_comb, self.data)
            if self.is_boostraping_calculation():
                self.results += boostraping_calculation(sample1=list(self._c[self.feature]),
                                                        sample2=list(self._a[self.feature]),
                                                        iteration=self._params['iteration'],
                                                        sample_size=self._params['sample_size'],
                                                        alpha=1-self._params['confidence_level'],
                                                        dist=self.data_distribution)
            else:
                self.results += bayesian_approach(sample1=list(self._c[self.feature]),
                                                  sample2=list(self._a[self.feature]), dist=self.data_distribution)
        self.results = DataFrame(self.results)

    def check_for_time_period(self):
        return True if self.time_period is None else False

    def get_time_period_of_time_line(self):
        if not self.check_for_time_period():
            self.time_indicator_values = sorted(self.data[self.time_period].unique())

    def run_test(self):
        try:
            self.f_w_data = self.data.query(self.get_query())
            if len(self.f_w_data) != 0:
                print("*" * 4, "AB TEST - ", self.get_query().replace(" and ", "; ").replace(" == ", " - "), "*" * 4)
                print("data size :", len(self.f_w_data))
                self.get_control_and_active_data()
                self.test_execute()
                self.test_decision()
        except Exception as e:
            print(e)

    def execute(self):
        print("time period :", self.time_period)
        self.decide_distribution()
        self.get_time_period_of_time_line()
        for self.comb in self.levels:
            if self.check_for_time_period():
                self.run_test()
            else:
                for self.tp in self.time_indicator_values:
                    print(self.tp)
                    self.run_test()









import datetime
import sys
import importlib
import os
import yaml
import json
import pickle
import datetime
import subprocess
import os
from os.path import join
import signal
import math
import random
import socket
import errno
import urllib
import time
from os import listdir
from os.path import dirname, join, abspath
import pandas as pd

try:
    from .configs import weekdays, conf, boostrap_ratio, web_port_default
except Exception as e:
    from configs import weekdays, conf, boostrap_ratio, web_port_default


def callfunc(my_file):
    pathname, filename = os.path.split(my_file)
    sys.path.append(os.path.abspath(pathname))
    modname = os.path.splitext(filename)[0]
    my_mod = importlib.import_module(modname)
    return my_mod


def get_running_pids(process_name, argument=None):
    pids = []
    p = subprocess.Popen(['ps', '-A'], stdout=subprocess.PIPE)
    out, err = p.communicate()
    for line in out.splitlines():
        if process_name in line.decode('utf-8'):
            pid = int(line.decode('utf-8').split(None, 1)[0])
            if argument:
                if len(line.decode('utf-8').split(process_name)) != 0:
                    args = line.decode('utf-8').split(process_name)[1].split(None, 1)
                    if str(argument) in args:
                        print("running job :", line.decode('utf-8'))
                        pids.append(pid)
            else:
                print("running job :", line.decode('utf-8'))
                pids.append(pid)
        else:
            print("no initialized ", argument, " is detected")
    return pids


def kill_process_with_name(process_name, argument=None):
    pids = get_running_pids(process_name, argument=argument)
    if len(pids) != 0:
        for pid in pids:
            os.kill(pid, signal.SIGKILL)
    else:
        print("no running jobs")


def is_port_in_use(port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    use = False
    try:
        s.bind(("0.0.0.0", port))
    except socket.error as e:
        if e.errno == errno.EADDRINUSE:
            use = True
    return use


def read_yaml(directory, filename):
    with open(join(directory, "", filename)) as file:
        docs = yaml.full_load(file)
    return docs


def write_yaml(directory, filename, data, ignoring_aliases=False):
    if ignoring_aliases:
        yaml.Dumper.ignore_aliases = lambda *args : True

    with open(join(directory, "", filename), 'w') as file:
        if ignoring_aliases:
            yaml.dump(data, file, default_flow_style=False)
        else:
            yaml.dump(data, file)


def read_write_to_json(directory, filename, data, is_writing):
    if is_writing:
        with open(join(directory, "", filename), 'w') as file:
            json.dump(data, file)
    else:
        with open(join(directory, "", filename), "r") as file:
            data = json.loads(file.read())
        return data


def model_from_to_pkl(directory=None, path=None, model=None, is_writing=False):
    if is_writing:
        with open(join(directory, "", path), "wb") as f:
            pickle.dump(model, f)
    else:
        with open(join(directory, "", path), 'rb') as f:
            model = pickle.load(f)
        return model


def get_col(c1, c2):
    if c2 == 1:
        return c1
    else:
        return c1 + '_' + str(c2)


def split_groups(groups):
    s_groups = []
    if groups not in [None, 'None']:
        if "+" not in groups:
            s_groups = [groups]
        else:
            s_groups = groups.split("+")
    return s_groups


def split_test_groups(group, data):
    return sorted(list(data[group].unique()))[-1]


def get_day_part(hour):
    if 0 <= hour < 7:
        return 'night'
    if 7 <= hour < 12:
        return 'morning'
    if 12 <= hour < 17:
        return 'afternoon'
    if 17 <= hour < 21:
        return 'evening1'
    if 21 <= hour < 24:
        return 'evening2'


def date_part(date, part):
    if part == 'year':
        return date.year
    if part == 'quarter':
        return get_quarter(date)
    if part == 'month':
        return date.month
    if part == 'week':
        return date.isocalendar()[1]
    if part == 'week_part':
        return 1 if date.isoweekday() in [6, 7] else 0
    if part == 'week_day':
        return date.isoweekday()
    if part == 'day_part':
        return get_day_part(date.hour)
    if part == 'hour':
        return date.hour
    if part == 'min':
        return date.min
    if part == 'second':
        return date.second
    if part == 'day':
        return str(date)[0:10]


def get_quarter(d):
    return "Q%d_%d" % (math.ceil(d.month/3), d.year)


def model_path(comb, group, model):
    return "_".join(["_".join([str(i[0]) + "*" + str(i[1]) for i in zip(group, comb)]), model]) + ".json"


def convert_date(date=None, time=None, str=False):
    if date not in ['', None]:
        date_merged = date + ' ' + time if time != '' else date
        if len(date_merged) == 0:
            format_str = '%Y-%m-%d'
        if len(date_merged) == 16:
            format_str = '%Y-%m-%d %H:%M'
        if len(date_merged) > 16:
            format_str = '%Y-%m-%d %H:%M:%S.%f'
        if not str:
            date_merged = datetime.datetime.strptime(date_merged, format_str)
    else:
        date_merged = datetime.datetime.now() + datetime.timedelta(minutes=2)
    return date_merged


def convert_date(x):
    return datetime.datetime.strptime(str(x), "%Y-%m-%d %H:%M:%S")


def convert_str_to_day(x):
    try:
        return datetime.datetime.strptime(str(x)[0:10], "%Y-%m-%d")
    except Exception as e:
        return None


def get_split_date(period=None, dates=None, params=None):
    model_infos = read_yaml(conf('model_main_path'), 'model_configuration.yaml')['infos']
    if period is not None:
        today = max(dates) if dates is not None else convert_date(model_infos['max_date'])
        isoweekday = {weekdays[i]: i for i in range(7)}
        day_diffs = {"Daily": 1, 'Weekly': 7, 'Every 2 Weeks': 14, 'Monthly': 30}
        hour_diffs = {"hourly": datetime.timedelta(hours=1), 'Every Min': datetime.timedelta(minutes=1), 'Every Second': datetime.timedelta(seconds=1)}
        if period in weekdays:
            last_isoweekday = lambda x: today + datetime.timedelta(days=-today.weekday()) + datetime.timedelta(days=x)
            split_date = last_isoweekday(isoweekday[period])
        if period in list(day_diffs.keys()):
            split_date = today - datetime.timedelta(days=day_diffs[period])
        if period in list(hour_diffs.keys()):
            split_date = today - hour_diffs[period]
    else:
        if dates is not None:
            split_date = get_ratio_of_date(max(dates), min(dates), params['split_ratio'])
        else:
            max_date, min_date = convert_date(model_infos['max_date']), convert_date(model_infos['min_date'])
            split_date = get_ratio_of_date(max_date, min_date, params['split_ratio'])
    return split_date


def get_ratio_of_date(max_date, min_date, ratio):
    range_dates = (max_date - min_date).total_seconds()
    return min_date + datetime.timedelta(seconds=int(range_dates * ratio))


def get_query_date(job, period=None, dates=None, params=None):
    date = None
    if job == 'prediction':
        date = get_split_date(period=period, dates=dates, params=params)
    return date


def url_string(value, res=False):
    if value is not None:
        if res:
            return value.replace("\r", " ").replace("\n", " ").replace(" ", "+")
        else:
            return value.replace("+", " ")
    else:
        return None


def request_url(url, params=None):
    if params is not None:
        url += '?'
        for p in params:
            url += p + '=' + url_string(str(params[p]), res=True) + '&'
    response = 404
    while response != 200:
        try:
            res = urllib.request.urlopen(url)
            response = res.code
        except Exception as e:
            print(e)
        time.sleep(2)


def convert_feature(value):
    try:
        if value == value:
            return value
        else:
            return None
    except Exception as e:
        print(e)
        return None


def get_sample_size(size):
    if size > 30:
        if size >= 2000:
            return 1000
        else:
            return int(size * boostrap_ratio)
    else:
        return 5


def get_residuals(residuals, ratio):
    if len(residuals) > 10000:
        return random.sample(residuals, 10000)
    else:
        return random.sample(residuals, len(residuals) * ratio)


convert_date_to_day = lambda x:  datetime.datetime.strptime(str(x)[0:10], "%Y-%m-%d %H:%M:%S")


def show_chart(df, x, y1, y2, is_bar_chart):
    """
    plots line chart with to lines
    params: df; x, y1, y2 fields included, x: x axis of data set, y1, y2 and y3; y axis of data sets
    params: is_bar_chart if True shows Bar chart, is_sorted: if True sorts y1, y2 and y3
    return: return multi dimensional line chart or bar chart on plotly
    """
    import plotly.graph_objs as go
    import plotly.offline as offline
    offline.init_notebook_mode()

    chart = go.Bar if is_bar_chart else go.Scatter
    marker = {'size': 15, 'opacity': 0.5, 'line': {'width': 0.5, 'color': 'white'}}
    x = df[x]
    _y1, _y2 = df[y1], df[y2]
    trace = []
    names = [y1, y2]
    counter = 0
    for _y in [_y1, _y2]:
        if not is_bar_chart:
            trace.append(chart(x=x, y=_y, mode='lines+markers', name=names[counter], marker=marker))
        else:
            trace.append(chart(x=x, y=_y, name=names[counter]))
        counter += 1
    offline.iplot(trace)


def get_results(date_col):
    results = []
    for f in listdir(dirname(join(conf('data_main_path'), ""))):
        f_splits = f.split(conf('result_file'))
        if f_splits[0] == "":
            results += pd.read_csv(join(conf('data_main_path'), "", f)).to_dict('results')
    results = pd.DataFrame(results)
    if len(results) >= 1000:
        results = results.sort_values(by=date_col, ascending=True)[-1000:]
    return results


def find_web_port():
    web_port = web_port_default
    while is_port_in_use(web_port):
        web_port += 1
    return web_port


def convert_dt_str(date, replace=True):
    if replace:
        return str(date)[0:10].replace("-", "")
    else:
        return str(date)[0:10]


def check_result_data_exits(path):
    file = []
    for f in listdir(dirname(join(path, ""))):
        if len(f.split("_")) != 1:
            if f.split("_")[-1] == "results.csv":
                file += [f]
    return file


def get_result_file_name(path, date, time_period=None):
    if time_period is None:
        if date is not None:
            return join(path, convert_dt_str(date, replace=True) + "_results.csv")
        else:
            return join(path,  "results.csv")
    else:
        if date is not None:
            return join(path, date[0:19].replace("-", "").replace(" ", "") + "_results.csv")
        else:
            return join(path,  time_period + "_results.csv")


def get_folder_path():
    return abspath(__file__).split("utils.py")[0]








from os.path import join
import subprocess

try:
    from main import main
    from data_access import GetData
    from utils import get_folder_path, write_yaml, read_yaml
    from configs import conf
except Exception as e:
    from .main import main
    from .data_access import GetData
    from .utils import get_folder_path, write_yaml, read_yaml
    from .configs import conf


class ABTest:
    """
    test_groups:        column of the data which represents  A - B Test of groups.
                        It  is a column name from the data.
                        AB test runs as control  - active group name related to columns of unique values.
                        This column has to 2 unique values which shows us the test groups

    groups:             column of the data which represents  individual groups for Testing.
                        AB Testing will be applied for each Groups which are unique value of groups column in data.

    date:               This parameter represent calculating dates.
                        If there is time schedule for AB Test, according to date column condition will be
                        "< = date". All data before the given date (given date is included)
                        must be collected for test results.

    feature:            Represents testing values of Test.
                        Test calculation will be applied according the feature column

    data_source:        AWS RedShift, BigQuery, PostgreSQL, csv, json files can be connected to system
                        E.g.
                        {"data_source": ..., "db": ..., "password": ..., "port": ..., "server": ..., "user": ...}

    data_query_path:    if there is file for data importing;
                            must be the path (e.g /../.../ab_test_raw_data.csv)
                        if there is ac- connection such as PostgreSQL / BigQuery
                            query must at the format "SELECT++++*+++FROM++ab_test_table_+++"

    time_indicator      This can only be applied with date. It can be hour, day, week, week_part, quarter, year, month.
                        Individually time indicator checks the date part is significantly
                        a individual group for data set or not.
                        If it is uses time_indicator as a  group

    export_path        exporting the results data to. only path is enough for importing data with .csv format.
    """
    def __init__(self,
                 test_groups,
                 groups=None,
                 date=None,
                 feature=None,
                 data_source=None,
                 data_query_path=None,
                 time_period=None,
                 time_indicator=None,
                 time_schedule=None,
                 export_path=None,
                 connector=None,
                 confidence_level=None,
                 boostrap_sample_ratio=None,
                 boostrap_iteration=None):
        self.test_groups = test_groups
        self.groups = groups
        self.date = date
        self.feature = feature
        self.data_source = data_source
        self.data_query_path = data_query_path
        self.time_period = time_period
        self.time_indicator = time_indicator
        self.time_schedule = time_schedule
        self.export_path = export_path
        self.connector = connector
        self.confidence_level = confidence_level
        self.boostrap_sample_ratio = boostrap_sample_ratio
        self.boostrap_iteration = boostrap_iteration
        self.arguments = {"test_groups": test_groups,
                          "groups": groups,
                          "date": date,
                          "feature": feature,
                          "data_source": data_source,
                          "data_query_path": data_query_path,
                          "time_period": time_period,
                          "time_indicator": time_indicator,
                          "export_path": export_path,
                          "parameters": None}
        self.arg_terminal = {"test_groups": "TG",
                             "groups": "G",
                             "date": "D",
                             "feature": "F",
                             "data_source":  "DS",
                             "data_query_path": "DQP",
                             "time_period": "TP",
                             "time_indicator": "TI" ,"export_path": "EP", "parameters": "P"}
        self.args_str = ""
        self.ab_test = None
        self.path = get_folder_path()
        self.params = None

    def arguments_for_subprocess(self):
        """
        when scheduling the AB Test, it is running from shell format EX: "python .../../../scheduler_service.py
        these arguments combined with string format to call with subprocess library.
        """
        for arg in self.arg_terminal:
            self.args_str += "-" + self.arg_terminal[arg] + " " + self.arguments[arg]

    def get_connector(self):
        """
       query_string_change İf data
        """
        config = conf('config')
        try:
            if self.data_source not in ["csv", "json"]:
                for i in config['db_connection']:
                    print(i)
                    if i != 'data_source':
                        config['db_connection'][i] = self.connector[i]
                    else:
                        config['db_connection']['data_source'] = self.data_source
            write_yaml(join(self.path, "docs"), "configs.yaml", config, ignoring_aliases=False)
            source = GetData(data_source=self.data_source,
                             date=self.date,
                             data_query_path=self.data_query_path,
                             time_indicator=self.time_indicator,
                             feature=self.feature)
            source.get_connection()
            return True
        except Exception as e:
            print(e)
            if self.data_source not in ["csv", "json"]:
                for i in config['db_connection']:
                    if i is not 'data_source':
                        config['db_connection'][i] = None
                    else:
                        config['db_connection']['data_source'] = self.data_source
            write_yaml(join(self.path, "docs"), "configs.yaml", config, ignoring_aliases=False)
            return False

    def query_string_change(self):
        if self.data_source in ['mysql', 'postgresql', 'awsredshift', 'googlebigquery']:
            self.data_query_path = self.data_query_path.replace("\r", " ").replace("\n", " ").replace(" ", "+")

    def check_for_time_period(self):
        if self.time_period is None:
            return True
        else:
            if self.time_period in ["day", "year", "month", "week", "week_day",
                                    "hour", "quarter", "week_part", "day_part"]:
                return True
            else: return False

    def assign_test_parameters(self, param, param_name):
        print(param, param_name)
        if param is not None:
            print("yessss::::SA:DAS:DASD")
            for i in self.params:
                if type(param) == list:
                    if len([i for i in param if 0 < i < 1]) != 0:
                        self.params[i][param_name] = "_".join([str(i) for i in param if 0 < i < 1])
                else:
                    if 0 < param < 1:
                        self.params[i][param_name] = str(param)

    def check_for_test_parameters(self):
        if self.confidence_level is not None or self.boostrap_sample_ratio is not None:
            self.params = read_yaml(join(self.path, "docs"), "test_parameters.yaml")['test_parameters']
            for _p in [(self.confidence_level, "confidence_level"),
                       (self.boostrap_sample_ratio, "sample_size"),
                       (self.boostrap_iteration, "iteration")]:
                self.assign_test_parameters(param=_p[0], param_name=_p[1])
            self.arguments["parameters"] = self.params

    def ab_test_init(self):
        self.check_for_test_parameters()
        self.query_string_change()
        if self.get_connector():
            if self.check_for_time_period():
                self.ab_test = main(**self.arguments)
            else:
                print("optional time periods are :")
                print("year", "month", "week", "week_day", "hour", "quarter", "week_part", "day_part")
        else:
            print("pls check for data source connection / path / query.")

    def schedule_test(self):
        if self.get_connector():
            cmd = "python " + self.path + " " + self.args_str
            result = subprocess.Popen(cmd, shell=True)
        else:
            print("pls check for data source connection / path / query.")

    def show_dashboard(self):
        """
        if you are running dashboard make sure you have assigned export_path.
        """




import sys
from os.path import abspath
import argparse

try:
    from ab_test import Test
except Exception as e:
    from .ab_test import Test

try:
    from utils import kill_process_with_name, url_string, get_result_file_name
except Exception as e:
    from .utils import kill_process_with_name, url_string, get_result_file_name


def main(test_groups, groups=None, date=None, feature=None, data_source=None,
         data_query_path=None, time_period=None, time_indicator=None, export_path=None, parameters=None):
    print("received :", {'test_groups': test_groups, 'groups': groups, 'date': date,
                         'feature': feature, 'data_source': data_source,
                         'data_query_path': data_query_path, 'time_period': time_period,
                         "time_indicator": time_indicator, "export_path": export_path, "parameters": parameters}
          )
    print(Test.__init__)
    ab_test = Test(test_groups=test_groups, groups=groups, date=date, feature=feature,
                     data_source=data_source, data_query_path=url_string(data_query_path),
                     time_period=time_period, time_indicator=time_indicator, export_path=export_path, parameters=parameters)
    ab_test.execute()
    print("Done!!")
    if export_path is None:
        return ab_test
    else:
        print("exporting path :", export_path)
        print(ab_test.final_results.head())
        ab_test.final_results.to_csv(get_result_file_name(export_path, date, time_period), index=False)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-TG", "--test_groups", type=str,
                        help="""
                        column of the data which represents  A - B Test of groups. 
                        It  is a column name from the data.
                        AB test runs as control  - active group name related to columns of unique values.
                        This column has to 2 unique values which shows us the test groups
                        """,
                        required=True)
    parser.add_argument("-G", "--groups", type=str,
                        help="""
                        column of the data which represents  individual groups for Testing. 
                        AB Testing will be applied for each Groups which are unique value of groups column in data.
                        """,
                        required=True)
    parser.add_argument("-D", "--date", type=str,
                        help="""
                        This parameter represent calulating dates. 
                        If there is time schedule for AB Test, according to date column condintion will be
                        "< = date". All data before the given date (given date is included) 
                        must be collected for test results.
                        """,
                        )
    parser.add_argument("-F", "--feature", type=str,
                        help="""
                        Represents testing values of Test. Test calculation will be applied according the feature column
                        """,
                        required=True)
    parser.add_argument("-DS", "--data_source", type=str,
                        help="""
                        AWS RedShift, BigQuery, PostgreSQL, csv, json files can be connected to system
                        """,
                        required=True)
    parser.add_argument("-DQP", "--data_query_path", type=str,
                        help="""
                        if there is file for data importing;
                            must be the path (e.g /../.../ab_test_raw_data.csv)
                        if there is ac- connection such as PostgreSQL / BigQuery
                            query must at the format "SELECT++++*+++FROM++ab_test_table_+++"
                        """,
                        required=True)
    parser.add_argument("-TI", "--time_indicator", type=str,
                        help="""
                        This can only be applied with date. It can be hour, day, week, week_part, quarter, year, month.
                        Individually time indicator checks the date part is significantly 
                        a individual group for data set or not.
                        If it is uses time_indicator as a  group
                        """,
                        )
    parser.add_argument("-TP", "--time_period", type=str,
                        help="""
                        This shows us to the time period if AB Test is running sequentially
                        """,
                        )
    parser.add_argument("-EP", "--export_path", type=str,
                        help="""
                        This shows us to the time period if AB Test is running sequentially
                        """,
                        )

    parser.add_argument("-P", "--parameters", type=str,
                        help="""
                        send test parameter (e.g Confidence interval)
                        """,
                        )
    arguments = parser.parse_args()
    args = {'test_groups': arguments.test_groups, 'groups': arguments.groups, 'date': arguments.date,
            'feature': arguments.feature, 'data_source': arguments.data_source,
            'data_query_path': arguments.data_query_path, 'time_period': arguments.time_period,
            'time_indicator': arguments.time_indicator,
            'export_path': arguments.export_path, "parameters":arguments.parameters}
    print(args)
    main(**args)



import datetime
import yaml
from os.path import join, abspath
from dateutil.parser import parse


init_directory = abspath(__file__).split("configs.py")[0]  # "/".join(abspath(__file__).split("/")[:-1]) + "/"
init_directory2 = abspath("")


def get_directory(path):
    directory, web, web_host = abspath(__file__).split("configs.py")[0], 7000, '127.0.0.1'
    with open(join(path, "instance.yaml")) as file:
        instances = yaml.full_load(file)
    active_ins = []
    for ins in instances['instances']:
        if ins['active'] is True and init_directory2 == ins['absolute_path']:
            active_ins.append(ins)
    if len(active_ins) == 0:
        directory = init_directory
        web = 7070
    if len(active_ins) == 1:
        active = active_ins[0]
        directory = active['directory']
        web = active['web']
        web_host = active['web_host']
    if len(active_ins) > 1:
        now = datetime.datetime.now()
        actives = list(map(lambda x: ((now - x['start_date']).total_seconds(), x['directory'], x['web'], x['web_host']), active_ins))
        directory = sorted(actives)[0][1]
        web = sorted(actives)[0][2]
        web_host = sorted(actives)[0][3]
    return directory, web, web_host


def read_config(directory):

    with open(join(directory, "docs", "configs.yaml")) as file:
        config = yaml.full_load(file)
    return config


def read_params(directory):
    with open(join(directory, "docs", "test_parameters.yaml")) as file:
        config = yaml.full_load(file)
    return config


def conf(var):
    config = read_config(abspath(__file__).split("configs.py")[0])
    return {
             'config': {c: config[c] for c in config
                        if c not in
                        ['data_main_path', 'model_main_path', 'log_main_path', 'docs_main_path', 'folder_name']}
    }[var]


def hyper_conf(var):
    config = read_params(abspath(__file__).split("configs.py")[0])
    return {'distribution_parameters': config['test_parameters'],
            'normal': config['test_parameters']['normal'],
            'beta': config['test_parameters']['beta']}[var]


alpha = 0.01
iteration = 30
boostrap_ratio = 0.5
time_dimensions = ['year', 'quarter', 'month', 'week', 'week_part', 'week_day', 'day_part', 'hour', 'min', 'second']
weekdays = ['Mondays', 'Tuesdays', 'Wednesdays', 'Thursdays', 'Fridays', 'Saturdays', 'Sundays']
folder_name = 'anomaly_detection_framework'
web_port_default = 7002
day_of_year = 366
time_indicator_accept_threshold = {
    'threshold': 0.9, 'accept_ratio_value': 0.5
}
s_size_ratio = 0.6
normal_dist_confidence_intervals = [.9, .99, .01]
descriptive_columns = ['mean1', 'mean2', 'size1', 'size2', 'var1', 'var2']


from itertools import product, combinations
import numpy as np
import pandas as pd
import random
from math import sqrt
from scipy import stats
from dateutil.parser import parse

try:
    from data_access import GetData
    from utils import *
    from configs import time_dimensions, day_of_year, time_indicator_accept_threshold, s_size_ratio
except Exception as e:
    from .data_access import GetData
    from .utils import *
    from .configs import time_dimensions, day_of_year, time_indicator_accept_threshold, s_size_ratio


def data_manipulation(date, time_indicator, feature, data_source, groups, data_query_path, time_period):
    data_process = GetData(data_source=data_source,
                           data_query_path=data_query_path,
                           time_indicator=time_indicator,
                           feature=feature,
                           date=date)
    data_process.data_execute()
    print("data size :", len(data_process.data))
    print("remove null values ..")
    data_process.data = data_process.data[data_process.data[feature] == data_process.data[feature]]
    print("check for time part data ..")
    data, groups = data_process.data, split_groups(groups)
    if time_indicator is not None:
        date_features = TimePartFeatures(job='train',
                                         data=data_process.data,
                                         time_indicator=time_indicator,
                                         groups=groups,
                                         feature=feature)
        date_features.date_dimension_deciding()
        data, groups = date_features.data, date_features.groups
        if time_period is not None:
            data[time_period] = data[time_indicator].apply(lambda x: date_part(str(x), time_period))
    return data, groups


class TimePartFeatures:
    def __init__(self, job=None, data=None, time_indicator=None, groups=None, feature=None):
        self.job = job
        self.data = data
        self._data = data
        self.time_indicator = time_indicator
        self.groups = groups
        self.time_groups = None
        self.date = time_indicator
        self.feature = feature
        self.time_dimensions, self.smallest_time_indicator = smallest_time_part(list(self.data[self.time_indicator]))
        self.time_dimensions_accept = {d: False for d in self.time_dimensions}
        self.threshold = time_indicator_accept_threshold['threshold']
        self.accept_ratio_value = time_indicator_accept_threshold['accept_ratio_value']
        self.data_distribution = 'normal'
        self.results = pd.DataFrame()
        self.h0_accept_ratio = 0

    def remove_similar_time_dimensions(self, part):
        accept = False
        if part == 'year':
            if self.smallest_time_indicator not in ['week', 'week_part', 'week_day', 'day']:
                accept = True
        if part == 'quarter':
            if self.smallest_time_indicator not in ['week', 'week_part', 'week_day', 'day']:
                if not self.time_dimensions_accept['year']:
                    accept = True
        if part == 'month':
            if self.smallest_time_indicator not in ['week', 'week_part', 'week_day', 'day']:
                if not self.time_dimensions_accept['quarter']:
                    accept = True
        if part == 'week':
            if self.smallest_time_indicator in ['hour', 'min', 'second']:
                if len([1 for p in ['year', 'quarter', 'month'] if self.time_dimensions_accept[p]]) == 0:
                    accept = True
        if part == 'week_part':
            if self.smallest_time_indicator in ['day', 'hour']:
                accept = True
        if part == 'week_day':
            if self.smallest_time_indicator in ['hour', 'min', 'second']:
                if not self.time_dimensions_accept['week_part']:
                    accept = True
        if part == 'day_part':
            if self.smallest_time_indicator in ['min', 'second']:
                accept = True
        if part == 'hour':
            if self.smallest_time_indicator == 'second':
                if not self.time_dimensions_accept['day_part']:
                    accept = True
        return accept

    def iteration_count(self, s1, s2):
        iter = int(min(len(s1), len(s2)) * 0.0001)
        if 1000 < min(len(s1), len(s2)) < 10000:
            iter = int(min(len(s1), len(s2)) * 0.01)
        if min(len(s1), len(s2)) < 1000:
            iter = int(min(len(s1), len(s2)) * 0.1)
        return iter

    def get_threshold(self, part):
        update_values = False
        if part == 'quarter':
            if self.time_dimensions_accept['year']:
                update_values = True
        if part == 'month':
            if self.time_dimensions_accept['year'] or self.time_dimensions_accept['quarter']:
                update_values = True
        if part == 'week':
            if len([1 for p in ['year', 'quarter', 'month'] if self.time_dimensions_accept[p]]) != 0:
                update_values = True

        if update_values:
            self.threshold = time_indicator_accept_threshold['threshold'] - 0.2
            self.accept_ratio_value = time_indicator_accept_threshold['accept_ratio_value'] + 0.2

    def decide_distribution(self):
        _unique = list(self.data[self.feature].unique())
        _type = type(_unique[0])
        # by default it is Normal distribution
        if _type != str:
            _min, _max = min(self.data[self.feature]), max(self.data[self.feature])
            if 0 <= _min < 1 and 0 < _max <= 1:
                self.data_distribution = 'beta'
        if len(_unique) == 2:
            self.data_distribution = 'binominal'
        if 2 < len(_unique) < 30:
            if _type == int:
                if min(self.data[self.feature]) >= 0:
                    self.data_distribution = 'poisson'
            if _type == str:
                self.data_distribution = 'poisson'
        print("Distribution :", self.data_distribution)

    def time_dimension_decision(self, part):
        self.decide_distribution()
        if self.remove_similar_time_dimensions(part):
            self.get_threshold(part=part)
            accept_count = 0
            combs = list(combinations(list(self.data[part].unique()), 2))
            for comb in combs:
                sample_1 = sampling(list(self._data[self._data[part] == comb[0]][self.feature]), sample_size=100000)
                sample_2 = sampling(list(self._data[self._data[part] == comb[1]][self.feature]), sample_size=100000)
                iter = self.iteration_count(sample_1, sample_2)
                self.results = boostraping_calculation(sample1=sample_1,
                                                                  sample2=sample_2,
                                                                  iteration=iter,
                                                                  sample_size=int(
                                                                      min(len(sample_1), len(sample_2)) * s_size_ratio),
                                                                  alpha=0.05, dist=self.data_distribution)
                self.results = pd.DataFrame(self.results)

                self.h0_accept_ratio = sum(self.results['h0_accept']) / len(self.results)
                accept_count += 1 if self.h0_accept_ratio < self.threshold else 0
            accept_ratio = len(combs) * self.accept_ratio_value  # 50%
            print("Time Part :", part, "Accept Treshold :", accept_ratio, "Accepted Count :", accept_count)
            return True if accept_count > accept_ratio else False

    def day_decision(self):
        return True if self.smallest_time_indicator in ['min', 'second'] else False

    def year_decision(self):
        return True if int(self.time_diff / 60 / 60 / 24) >= (day_of_year * 2) else False

    def quarter_decision(self):
        return True if int(self.time_diff / 60 / 60 / 24) >= (day_of_year * 1) else False

    def check_for_time_difference_ranges_for_accepting_time_part(self, part):
        decision = False
        if part == 'year':
            decision = self.year_decision()
        if part == 'quarter':
            decision = self.quarter_decision()
        if part == 'week_day':
            decision = self.day_decision()
        return decision

    def decide_timepart_of_group(self, part):
        print("*" * 5, "decision for time part :", part, "*" * 5)
        result = False
        (unique, counts) = np.unique(list(self.data[part]), return_counts=True)
        if len(unique) >= 2:
            if 1 not in counts:
                if part not in ['week_day', 'hour', 'min', 'second']:
                    if self.check_for_time_difference_ranges_for_accepting_time_part(part):
                        result = self.time_dimension_decision(part)
                else:
                    if part == 'week_day':
                        results = self.check_for_time_difference_ranges_for_accepting_time_part(part)
                    if self.smallest_time_indicator == 'second' and part == 'hour':
                        result = self.time_dimension_decision(part)
        print("result :", " INCLUDING" if result else "EXCLUDING")
        self.time_dimensions_accept[part] = result
        return result

    def calculate_date_parts(self):
        accepted_date_parts = []
        for t_dimension in self.time_dimensions:
            if t_dimension not in self.groups:
                self.data[t_dimension] = self.data[self.date].apply(lambda x: date_part(x, t_dimension))
                if self.decide_timepart_of_group(part=t_dimension):
                    accepted_date_parts.append(t_dimension)
        self.time_groups = accepted_date_parts
        self.groups += accepted_date_parts

    def date_dimension_deciding(self):
        if self.time_indicator is not None:
            self.time_diff = get_time_difference(list(self.data[self.time_indicator]))
            self.calculate_date_parts()


def smallest_time_part(dates):
    sample_dates = random.sample(dates, int(len(dates) * 0.8))
    smallest = False
    t_dimensions = list(reversed(time_dimensions))
    count = 0
    if len(np.unique(sample_dates)) != 1:
        while not smallest:
            (unique, counts) = np.unique([date_part(d, t_dimensions[count]) for d in sample_dates], return_counts=True)
            if len(unique) >= 2:
                smallest = True
                smallest_td = t_dimensions[count]
                break
            count += 1
        accepted_t_dimensions = list(reversed(t_dimensions[count + 1:]))
    else:
        accepted_t_dimensions, smallest_td = t_dimensions, t_dimensions[0]
    return accepted_t_dimensions, smallest_td  # smallest time indicator not included to time_dimensions


def get_time_difference(dates):
    return (max(dates) - min(dates)).total_seconds()


def sampling(sample, sample_size):
    if len(sample) <= sample_size:
        return sample
    else:
        return random.sample(sample, sample_size)


def get_sample_size(ratio, sample):
    return int(ratio * len(sample))


def calculate_lambda(values):
    _unique, counts = np.unique(values, return_counts=True)
    return sum(list(map(lambda x: x[0] * x[1], zip(_unique, counts)))) / len(values)


def boostraping_calculation(sample1, sample2, iteration, sample_size, alpha, dist):
    """
    Randomly selecting samples from two population on each iteration.
    Iteratively independent test are applied each randomly selected samples
    :param sample1: list of values related to sample 1
    :param sample2: list of values related to sample 2
    :param iteration: numeber of iteration. It is better to asssign higher values.
    :param sample_size: number of sample when randomly sampled from each data set.
                        Make sure this parameters bigger than both sample of size
    :param alpha: Confidence level
    :param two_sample: Is it related to Two Sample Test or not.
    :return: HO_accept ratio: num of accepted testes / iteration
             result data set: each iteration of test outputs of pandas data frame
    """
    if dist == 'poisson':
        _unique, counts = np.unique(sample1+sample2, return_counts=True)
        _lookups = {i[0][1]: i[1] for i in list(zip(reversed(sorted(list(zip(counts, _unique)))), range(len(_unique))))}
    pval_list, h0_accept_count, test_parameters_list= [], 0, []
    for i in range(iteration):
        try:
            d = {'sample_ratio': sample_size, 'confidence_level': alpha, 'h0_accept': 0}
            d['size1'] = get_sample_size(d['sample_ratio'], sample1)
            d['size2'] = get_sample_size(d['sample_ratio'], sample1)
            # random.sample(sample1, sample_size)  # randomly picking samples from sample 1
            random1 = sampling(sample=sample1,
                               sample_size=d['size1'])
            # random.sample(sample2, sample_size)  # randomly picking samples from sample 2
            random2 = sampling(sample=sample2,
                               sample_size=d['size2'])
            if dist == 'normal':
                d['mean1'], d['mean2'] = np.mean(random1), np.mean(random2)
                d['var1'], d['var2'] = np.var(random1), np.var(random2)
            if dist == 'binominal':
                true_value = sorted(np.unique(random1).tolist())[-1]
                d['true_value1'] = len(list(filter(lambda x: x == true_value, random1)))
                d['mean1'], d['mean2'] = d['true_value1'] / d['size1'], d['true_value2'] / d['size2']
                d['true_value2'] = d['mean2'] * d['size1']
                d['var1'] = d['mean1'] * (1 - d['mean1']) * d['size1']
                d['var2'] = d['mean2'] * (1 - d['mean2']) * d['size2']
            if dist == 'poisson':
                random1, random2 = list(map(lambda x: _lookups[x], random1)), list(map(lambda x: _lookups[x], random2))
                d['mean1'], d['mean2'] = calculate_lambda(random1), calculate_lambda(random2)
                d['var1'], d['var2'] = d['mean1'], d['mean2']
            if dist not in ['binominal', 'poisson']:
                d['var1'], d['var2'] = np.var(random1), np.var(random2)
                d['mean1'], d['mean2'] = np.mean(random1), np.mean(random2)
            if dist == 'normal':
                d['pval'], d['confidence_intervals'], hypotheses_accept = calculate_t_test(d['mean1'], d['mean2'],
                                                                                           d['var1'], d['var2'],
                                                                                           d['size1'], d['size2'],
                                                                                           d['confidence_level'])
            if dist == 'beta':
                diff_values = list((map(lambda x: x[0] - x[1], zip(random1, random2))))
                d['diff_mean'], d['diff_var'] = np.mean(diff_values), np.var(diff_values)
                d['pval'], d['confidence_intervals'], hypotheses_accept = calculate_beta_test(d['diff_mean'],
                                                                                              d['diff_var'],
                                                                                              d['confidence_level'])
            if dist == 'binominal':
                d['pval'], d['confidence_intervals'], hypotheses_accept = calculate_binomial_test(d['mean1'],
                                                                                                  d['true_value2'],
                                                                                                  d['size1'],
                                                                                                  d['confidence_level'])
            if dist == 'poisson':
                d['pval'], d['confidence_intervals'], hypotheses_accept = calculate_poisson_test(d['mean1'],
                                                                                                  d['mean2'],
                                                                                                  d['confidence_level'])
            d['h0_accept'] += 1 if hypotheses_accept == 'HO ACCEPTED!' else 0
            test_parameters_list.append(d)
        except Exception as e:
            print(e)
    return test_parameters_list


def calculate_t_test(mean1, mean2, var1, var2, n1, n2, alpha):
    """
    It Test according to T- Distribution of calculations
    There are two main Hypotheses Test are able to run. Two Sample Two Tail Student-T Test, One Sample Student-T-Test
    In order to test two main parameters are needed Mean and Variance, T~(M1, Var)
    :param mean1: Mean of Sample 1
    :param mean2: Mean of Sample 2. If one sample T-Test assign None
    :param var1: Variance of Sample 1.
    :param var2: Variance of Sample 2. If one sample T-Test assign None
    :param n1: sample 1 size
    :param n2: sample 2 size. If one sample T-Test assign None
    :param alpha: Confidence level
    :param two_sample: Boolean; True - False
    :return: returns p-value of test, confidence interval of test, H0 Accepted!! or H0 REjected!!
    """
    # Two Sample T Test (M0 == M1) (Two Tails)
    try:
        t = (mean1 - mean2) / sqrt((var1 / n1) + (var2 / n2))  # t statistic calculation for two sample
        df = n1 + n2 - 2  # degree of freedom for two sample t - set
        pval = 1 - stats.t.sf(np.abs(t), df) * 2  # two-sided pvalue = Prob(abs(t)>tt) # p - value
        cv = stats.t.ppf(1 - (alpha / 2), df)
        standart_error = cv * sqrt((var1 / n1) + (var2 / n2))
        confidence_intervals = [abs(mean1 - mean2) - standart_error, abs(mean1 - mean2) + standart_error, standart_error]
        acception = 'HO REJECTED!' if pval < (alpha / 2) else 'HO ACCEPTED!'  # left tail
        acception = 'HO REJECTED!' if pval > 1 - (alpha / 2) else 'HO ACCEPTED!'  # right tail
        return pval, confidence_intervals, acception
    except Exception as e:
        return None, None, None


def calculate_beta_test(mean, var,  point):
    try:
        alpha = mean ** 2 * (1 - mean) / var - mean
        beta = alpha * (1 - mean) / mean
        p_value1 = stats.beta.cdf(float(point), alpha, beta)
        p_value2 = 1-float(stats.beta.cdf(float(point), alpha, beta))
        p_value = min(p_value1, p_value2) * 2
        confidence_intervals = [p_value1, p_value2]
        acception = 'HO ACCEPTED!' if ((1 - point) / 2) < p_value < (1 - ((1 - point) / 2)) else 'HO REJECTED!'
        return p_value, confidence_intervals, acception
    except Exception as e:
        return None, None, None


def calculate_binomial_test(mean, threatment_true_values, sample_size,  point):
    try:
        left_tail = stats.binom.ppf(1-point, sample_size, mean)
        right_tail = stats.binom.ppf(point, sample_size, mean)
        p_value = stats.binom.cdf(threatment_true_values, sample_size, mean)
        confidence_intervals = [left_tail, right_tail]
        acception = 'HO ACCEPTED!' if ((1 - point) / 2) < p_value < (1 - ((1 - point) / 2)) else 'HO REJECTED!'
        return p_value, confidence_intervals, acception
    except Exception as e:
        return None, None, None


def calculate_poisson_test(mu, lambda_value,  alpha):
    try:
        left_tail = stats.poisson.ppf(1 - alpha, mu)
        right_tail = stats.poisson.ppf(alpha, mu)
        p_value = stats.poisson.cdf(lambda_value, mu)
        confidence_intervals = [left_tail, right_tail]
        acception = 'HO ACCEPTED!' if ((1 - alpha) / 2) < p_value < (1 - ((1 - alpha) / 2)) else 'HO REJECTED!'
        return p_value, confidence_intervals, acception
    except Exception as e:
        print(e)


def bayesian_approach(sample1, sample2, dist):
    d = {}
    number_of_sample = max(len(sample1), len(sample2))
    a_control, b_control, a_val, b_val = get_p_values(sample1, sample2, dist)
    control_p_values = stats.beta.rvs(a_control, b_control, size=len(sample1))
    validation_p_values = stats.beta.rvs(a_val, b_val, size=len(sample2))
    sample_size = min(len(control_p_values), len(validation_p_values))
    wins = validation_p_values[:sample_size] > control_p_values[:sample_size]
    print(np.mean(wins))
    return np.mean(wins)


def bayesian_approach(sample1, sample2, dist):
    d = {'wins': None, 'a_control': 1, 'b_control': 1, 'a_val': 1, 'b_val': 1, 'accept_Ratio': 0}
    if dist == 'binominal':
        true_value = sorted(np.unique(sample1+sample2).tolist())[-1]
        sample1 = list(map(lambda x: 1 if x == true_value else 0, sample1))
        sample2 = list(map(lambda x: 1 if x == true_value else 0, sample2))
    if dist == 'normal':
        mean1, mean2, var1, var2, n = np.mean(sample1), np.mean(sample2), np.var(sample1), np.var(sample2), len(sample1)
        get_t_value = lambda x, mean, var: abs( (mean - x) / sqrt(var / n))
        sample1 = list(map(lambda x: stats.norm.sf(get_t_value(x, mean1, var1)), sample1))
        sample2 = list(map(lambda x: stats.norm.sf(get_t_value(x, mean2, var2)), sample2))
    if dist == 'poisson':
        lambda1, lambda2 = calculate_lambda(sample1), calculate_lambda(sample2)
        sample1 = list(map(lambda x: stats.poisson.sf(x, lambda1), sample1))
        sample2 = list(map(lambda x: stats.poisson.sf(x, lambda2), sample2))
        a_control, b_control, a_val, b_val = 1, 1, 1, 1
    for ind in list(range(max(len(sample1), len(sample2)))):
        # control set a, b updating
        try:
            d['a_control'] += sample1[ind]  # click
            d['b_control'] += abs(sample1[ind] - 1)  # non-click
        except:
            if ind + 1 == len(sample1):
                print("out of index")

        # validation set a, b updating
        try:
            d['a_val'] += sample2[ind]  # click
            d['b_val'] += abs(sample2[ind] - 1)  # non-click
        except:
            if ind + 1 == len(sample2):
                print("out of index")
    control_p_values = stats.beta.rvs(d['a_control'], d['b_control'], size=len(sample1))
    validation_p_values = stats.beta.rvs(d['a_val'], d['b_val'], size=len(sample2))
    sample_size = min(len(control_p_values), len(validation_p_values))
    d['wins'] = validation_p_values[:sample_size] > control_p_values[:sample_size]
    d['accept_Ratio'] = np.mean(d['wins'])
    d['wins'] = "_".join([str(i) for i in d['wins']])
    return [d]


def get_levels(data, groups):
    groups = [g for g in groups if g not in [None, '', 'none', 'null', 'None']]
    return list(product(*[list(data[data[g] == data[g]][g].unique()) for g in groups]))



import setuptools
from setuptools import find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="abtest",
    version="0.0.2",
    author="Caglan Akpinar",
    author_email="cakpinar23@gmail.com",
    description="allows to run AB Test for any problem, automatically decides which test must be applied and represents the results",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='A-B Test Control - Active Group Test AB Test via Bayesian Approach',
    packages= find_packages(exclude='__pycache__'),
    py_modules=['ab_test_platform/docs', 'ab_test_platform'],
    install_requires=[
                      "numpy >= 1.18.1",
                      "pandas >= 0.25.3",
                      "scipy >= 1.4.1 ",
                      "schedule >= 0.6.0",
                      "PyYAML",
                      "psycopg2-binary",
                      # "psycopg2 >= 2.8.5",
                      "python-dateutil >= 2.8.1",
                      "google-cloud-bigquery",
                      "mysql-connector-python",
                      "plotly >=  4.5.0",
                      "dash-html-components >= 1.0.2",
                      "dash-core-components >=  1.8.0",
                      "dash >= 1.9.0"
    ],
    url="https://github.com/caglanakpinar/abtp",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
)


import pandas as pd
from dateutil.parser import parse
from os.path import join
import os

try:
    from configs import conf
    from utils import read_yaml, read_write_to_json, convert_feature
except Exception as e:
    from .configs import conf
    from .utils import read_yaml, read_write_to_json, convert_feature


class GetData:
    def __init__(self, data_source=None, date=None, data_query_path=None, time_indicator=None, feature=None, test=None):
        self.data_source = data_source if data_source is not None else 'csv'
        self.data_query_path = data_query_path
        self.time_indicator = time_indicator
        self.feature = feature
        self.conn = None
        self.data = pd.DataFrame()
        self.nrows = test
        self.date = date
        self.query = data_query_path

    def get_connection(self):
        config = conf('config')
        if self.data_source in ['postgresql', 'awsredshift', 'mysql']:
            server, db, user, pw, port = str(config['db_connection']['server']), str(config['db_connection']['db']), \
                                         str(config['db_connection']['user']), str(config['db_connection']['password']),\
                                         int(config['db_connection']['port'])
        if self.data_source == 'mysql':
            from mysql import connector
            self.conn = connector.connect(host=server, database=db, user=user, password=pw)
        if self.data_source in ['postgresql', 'awsredshift']:
            import psycopg2
            self.conn = psycopg2.connect(user=user, password=pw, host=server, port=port, database=db)
        if self.data_source == 'googlebigquery':
            from google.cloud.bigquery.client import Client
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = join(conf('data_main_path'), "", config['db_connection']['db'])
            self.conn = Client()
        print("db connection is done!")

    def convert_date(self):
        try:
            self.data[self.time_indicator] = self.data[self.time_indicator].apply(lambda x: parse(str(x)))
        except Exception as e:
            print(e)

    def convert_feature(self):
        try:
            self.data[self.feature] = self.data[self.feature].apply(lambda x: convert_feature(x))
        except Exception as e:
            print(e)

    def check_data_with_filtering(self):
        if self.data_source in ('json', 'yaml', 'csv'):
            self.query = lambda x: x.query(self.time_indicator + " <= '" + str(self.date) + "'") if self.date is not None else x
        if self.data_source in ['mysql', 'postgresql', 'awsredshift', 'googlebigquery']:
            if self.date is not None and self.time_indicator is not None:
                print("SELECT * FROM (" + self.data_query_path + ") AS T WHERE T." + \
                             self.time_indicator + " <= '" + str(self.date) + "'   ")
                self.query = "SELECT * FROM (" + self.data_query_path + ") AS T WHERE T." + \
                             self.time_indicator + " <= '" + str(self.date) + "'   "

    def query_data_source(self):
        self.check_data_with_filtering()

        # import data via pandas
        if self.data_source in ['mysql', 'postgresql', 'awsredshift']:
            self.get_connection()
            self.data = pd.read_sql(self.query + " LIMIT " + str(self.nrows) if self.nrows else self.query, self.conn)

        # import data via google
        if self.data_source == 'googlebigquery':
            self.get_connection()
            self.data = self.conn.query(self.query + " LIMIT " + str(self.nrows) if self.nrows else self.query).to_dataframe()

        # import via pandas
        if self.data_source == 'csv':
            try:
                for sep in [',', ';', ':']:

                    self.data = pd.read_csv(filepath_or_buffer=join(conf('data_main_path'), self.data_query_path),
                                            error_bad_lines=False,
                                            encoding="ISO-8859-1",
                                            sep=sep,
                                            nrows=self.nrows)
                    if len(self.data.columns) > 1:
                        break
            except Exception as e:
                print(e)

        if self.data_source == 'json':
            self.data = read_write_to_json(conf('directory'), self.data_query_path, None, is_writing=False)

        if self.data_source == 'yaml':
            self.data = read_yaml(conf('data_main_path'), self.data_query_path)

        if self.data_source in ('json', 'yaml', 'csv'):
            self.data = self.query(self.data)

    def data_execute(self):
        self.query_data_source()
        self.convert_date()
        self.convert_feature()














import schedule
import datetime
import time
import argparse
from dateutil.parser import parse
from os.path import abspath, dirname
from os import listdir

try:
    from main import main
    from utils import date_part
except Exception as e:
    from .main import main
    from .utils import date_part


def get_schedule(time_period):
    if time_period not in ['minute', 'hour']:
        return {'Mondays': schedule.every().monday,
                'Tuesdays': schedule.every().tuesday,
                'Wednesdays': schedule.every().wednesday,
                'Thursdays': schedule.every().thursday,
                'Fridays': schedule.every().friday,
                'Saturdays': schedule.every().saturday,
                'Sundays': schedule.every().sunday,
                'Daily': schedule.every().day
                }[time_period]
    if time_period == 'week':
        return schedule.every().week
    if time_period == 'hour':
        return schedule.every(1).hours.at()


def update_date():
    return str(datetime.datetime.now())[0:10]


def run_ab_test():
    args['date'] = update_date()
    main(**args)


def create_job(ab_test_arguments, time_period):
    global args
    args = ab_test_arguments
    _sch = get_schedule(time_period)
    _sch.do(run_ab_test)
    while True:
        _sch.run_pending()
        time.sleep(1000)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-TS", "--time_schedule", type=str,
                        help="""
                        column of the data which represents  A - B Test of groups. 
                        It  is a column name from the data.
                        AB test runs as control  - active group name related to columns of unique values.
                        This column has to 2 unique values which shows us the test groups
                        """)
    parser.add_argument("-TG", "--test_groups", type=str,
                        help="""
                        column of the data which represents  A - B Test of groups. 
                        It  is a column name from the data.
                        AB test runs as control  - active group name related to columns of unique values.
                        This column has to 2 unique values which shows us the test groups
                        """,
                        required=True)
    parser.add_argument("-G", "--groups", type=str,
                        help="""
                        column of the data which represents  individual groups for Testing. 
                        AB Testing will be applied for each Groups which are unique value of groups column in data.
                        """,
                        required=True)
    parser.add_argument("-D", "--date", type=str,
                        help="""
                        This parameter represent calulating dates. 
                        If there is time schedule for AB Test, according to date column condintion will be
                        "< = date". All data before the given date (given date is included) 
                        must be collected for test results.
                        """,
                        )
    parser.add_argument("-F", "--feature", type=str,
                        help="""
                        Represents testing values of Test. Test calculation will be applied according the feature column
                        """,
                        required=True)
    parser.add_argument("-DS", "--data_source", type=str,
                        help="""
                        AWS RedShift, BigQuery, PostgreSQL, csv, json files can be connected to system
                        """,
                        required=True)
    parser.add_argument("-DQP", "--data_query_path", type=str,
                        help="""
                        if there is file for data importing;
                            must be the path (e.g /../.../ab_test_raw_data.csv)
                        if there is ac- connection such as PostgreSQL / BigQuery
                            query must at the format "SELECT++++*+++FROM++ab_test_table_+++"
                        """,
                        required=True)
    parser.add_argument("-TI", "--time_indicator", type=str,
                        help="""
                        This can only be applied with date. It can be hour, day, week, week_part, quarter, year, month.
                        Individually time indicator checks the date part is significantly 
                        a individual group for data set or not.
                        If it is uses time_indicator as a  group
                        """,
                        )
    parser.add_argument("-EP", "--export_path", type=str,
                        help="""
                        """,
                        )
    arguments = parser.parse_args()
    ab_test_arguments = {'test_groups': arguments.test_groups, 'groups': arguments.groups, 'date': arguments.date,
                         'feature': arguments.feature, 'data_source': arguments.data_source,
                         'data_query_path': arguments.data_query_path, 'time_period': arguments.time_period,
                         "export_path": arguments.export_path}

    create_job(ab_test_arguments, arguments.time_schedule)